/////////////////////////////////////////////////////////////////////
////////////////////////// Create constants /////////////////////////
/////////////////////////////////////////////////////////////////////

const pi2 = Math.PI * 2
const pi = Math.PI
const pi1_2 = pi/2

//Sizes
const margin = {
    top: 100,
    right: 100,
    bottom: 100,
    left: 100
}
const width = 1000 - margin.left - margin.right
const height = 1000 - margin.top - margin.bottom

//Puzzle dimensions
const color_stroke = "#0d0f1d"
const outside_radius = width/2
const cubby_radius = outside_radius * 0.925
const puzzle_radius = outside_radius * 0.6
const spirograph_radius = outside_radius * 0.8
const stroke_thickness = 2

//Line drawing functions
const line = d3.line()

//Spirograph - Farris Wheel
// https://walkingrandomly.com/?p=1471
const spirograph_settings = [
    { R: 0.5454545454545454, n: 1, ψ: 0 },
    { R: 0.2727272727272727, n: 7, ψ: 0 },
    { R: 0.18181818181818182, n: -17, ψ: 0.25 }
]
spirograph_settings.step_size = 0.0014285714285714286
spirograph_settings.steps = 701

const spirograph_path = line(calculateSpirograph(spirograph_settings))

const a = pi2 / 6 * 0.5

/////////////////////////////////////////////////////////////////////
////////////////////////// Create container /////////////////////////
/////////////////////////////////////////////////////////////////////

const svg = d3.select("body")
    .append("svg")
    .attr("width", width + margin.left + margin.right)
    .attr("height", height + margin.top + margin.bottom)
    .append("g")
    .attr("transform", `translate(${margin.left + width/2}, ${margin.top + height/2})`)

///////////////////////// Create the outside ////////////////////////

//Puzzle enclosure
const outside = svg.append("circle")
    .attr("r", outside_radius)
    .style("fill", "none")
    .style("stroke", "black")
    .style("stroke-width", 4)

//The spirograph etched out of the wood
const spirograph_etch_wood = svg.append("path")
    .attr("d", spirograph_path)
    .style("fill", "none")
    .style("stroke", "#000")
    .style("opacity", 0.8)
    .style("stroke-width", 4)

/////////////////////// Create the cubby holes //////////////////////

//Create hatch pattern to fill the cubby holes with
const defs = svg.append("defs")
const hatch_size = 10
const hatch_pattern = defs.append("pattern")
    .attr("id", "hatch")
    .attr("width", hatch_size)
    .attr("height", hatch_size)
    .attr("patternTransform", `rotate(${45})`)
    .attr("patternUnits", "userSpaceOnUse") //used to make the pattern independent of location of circle
//Add the actual pattern shape
hatch_pattern.append("rect")
    .attr("width", hatch_size)
    .attr("height", hatch_size)
    .style("fill", "#fff")
hatch_pattern.append("rect")
    .attr("width", hatch_size / 2)
    .attr("height", hatch_size)
    .style("fill", "#b6bbd2")
    .style("opacity", 0.4)


//Create a group for the cubby holes
const cubby_group = svg.append("g")
    .attr("class", "cubby-group")
    .style("stroke", "black")
    .style("stroke-width", 3)

//Draw the cubby holes
const offset = pi2 * 0.02
const inside_radius = cubby_radius * 0.8
cubby_group.selectAll(".cubby")
    .data(d3.range(4)) //make 4 holes
    .join("path")
    .attr("class", "cubby")
    .style("fill", "url(#hatch)")
    .attr("d", (d) => {
        let r = cubby_radius
        let a1 = pi2 * d / 6 - a + offset
        let a2 = pi2 * (d+1) / 6 - a - offset

        let x1 = r * Math.cos( a1 )
        let y1 = r * Math.sin( a1 )
        let x2 = r * Math.cos( a2 )
        let y2 = r * Math.sin( a2 )

        let path = `M${x1},${y1} A${r},${r} 0 0 1 ${x2},${y2}`

        let x = inside_radius * Math.cos( a2 )
        let y = inside_radius * Math.sin( a2 )
        path += ` L${x},${y}`

        x = inside_radius * Math.cos( a1 )
        y = inside_radius * Math.sin( a1 )
        path += ` L${x},${y}`

        path += ` L${x1},${y1}`

        return path + " Z"

    })

////////////// Create the inner hexagon puzzle outline //////////////

//Hexagon path for the outline
let hexagon_path = () => {
    let path = ""
    for(let i = 0; i < 7; i++) {
        let x1 = puzzle_radius * Math.cos( pi2 * i / 6 + a)
        let y1 = puzzle_radius * Math.sin( pi2 * i / 6 + a)

        if(i === 0) path = `M${x1},${y1}`
        else path += ` L${x1},${y1}`
    }
    path += " Z"
    return path
}//hexagon_path

//The hexagonal puzzle outline
const outside_puzzle = svg.append("path")
    .style("fill", "white")
    .style("fill-opacity", 0.8)
    .style("stroke", "black")
    .style("stroke-width", 5)
    .attr("d", hexagon_path)

//Clipping path of the hexagon
const clips = svg.append("g").attr("class", "clip-group")
clips.append("clipPath")
    .attr("id", "clip-area")
    .append("path")
    .attr("d", hexagon_path)

/////////////////// Create the inner puzzle pieces //////////////////

//Create center of hexagonal grid points
const SQRT3 = Math.sqrt(3)
const hex_radius = puzzle_radius * 0.2
let cols = Math.ceil((puzzle_radius * 2) / (hex_radius * SQRT3))
let rows = Math.ceil((puzzle_radius * 2) / (hex_radius * 1.5))
let cells = []
// console.log("Rows | Cols", rows, cols)
for (let i = 0; i < rows; i++) {
    for (let j = 0; j < cols; j++) {
        let x = hex_radius * j * SQRT3
        //Offset each uneven row by half of a "hex-width"
        if (i % 2 === 1) x += (hex_radius * SQRT3) / 2
        let y = hex_radius * i * 1.5
        cells.push([x,y])
    }//for j
}//for i


let hexagon_lines = []
let line_by_location = {}
//Create all the lines within the hexagonal grid
//Making sure that no lines overlap
cells.forEach(d => {
    //Create hexagon outline
    for(let i = 0; i < 6; i++) {
        let x1 = -puzzle_radius - 10 + d[0] + hex_radius * Math.cos( pi2 * i / 6 + a)
        let y1 = -puzzle_radius + d[1] + hex_radius * Math.sin( pi2 * i / 6 + a)
        let x2 = -puzzle_radius - 10 + d[0] + hex_radius * Math.cos( pi2 * (i+1) / 6 + a)
        let y2 = -puzzle_radius + d[1] + hex_radius * Math.sin( pi2 * (i+1) / 6 + a)

        let line = [[round(x1,2), round(y1,2)], [round(x2,2), round(y2,2)]]
        //Check if the line already exists
        if(!isLine(line)) {
            line_by_location[line[0] + "," + line[1]] = 1
            hexagon_lines.push(line)
        }
    }//for i

    //Create inner lines
    let x1 = -puzzle_radius - 10 + d[0]
    let y1 = -puzzle_radius + d[1]
    for(let i = 0; i < 6; i++) {
        let x2 = -puzzle_radius - 10 + d[0] + hex_radius * Math.cos( pi2 * (i+1) / 6 + a)
        let y2 = -puzzle_radius + d[1] + hex_radius * Math.sin( pi2 * (i+1) / 6 + a)

        let line = [[round(x1,2), round(y1,2)], [round(x2,2), round(y2,2)]]
        //Check if the line already exists
        if(!isLine(line)) {
            line_by_location[line[0] + "," + line[1]] = 1
            hexagon_lines.push(line)
        }
    }//for i
})//forEach

// //Now remove several lines randomly
// //Not perfect, because I don't want pieces going around corners
// // So I'll do this manually in Adobe Illustrator
// for(let i = 0; i < hexagon_lines.length * 1/3; i++){
//     hexagon_lines.splice(Math.floor(Math.random() * hexagon_lines.length), 1)
// }

//Create a group for all the "puzzle piece" edges
const puzzle_piece_group = svg.append("g")
    .attr("class", "puzzle-group")
    .attr("clip-path", "url(#clip-area)")
    .style("fill", "none")
    .style("stroke", "black")
    .style("stroke-width", 3)
//Add all the edges
puzzle_piece_group.selectAll(".hexagon-line")
    .data(hexagon_lines)
    .join("path")
    .attr("d", d => `M${d[0][0]},${d[0][1]} L${d[1][0]},${d[1][1]}`)

/////////////////////////////////////////////////////////////////////
//////////////////////// Spirograph Functions ///////////////////////
/////////////////////////////////////////////////////////////////////

function calculateSpirograph(wheels) {
    let line_data = []
    let θ, x, y
    for (i = 0; i < wheels.steps; i++) {
        θ = wheels.step_size * i * pi2
        x = 0
        y = 0
        //Loop over all the wheels
        wheels.forEach(w => {
            x += spirograph_radius * w.R * Math.cos(w.n * θ + pi2 * w.ψ)
            y += spirograph_radius * w.R * Math.sin(w.n * θ + pi2 * w.ψ)
        })//forEach
        line_data.push([round(x,2),round(y,2)])
    }//for i

    return line_data
}//function calculateSpirograph

/////////////////////////////////////////////////////////////////////
////////////////////////// Helper Functions /////////////////////////
/////////////////////////////////////////////////////////////////////

//Round to "precision" digits
function round(value, precision) {
    let multiplier = Math.pow(10, precision || 0)
    return Math.round(value * multiplier) / multiplier
}//function round

//Check if the line already exists
function isLine(line) {
    return line_by_location[line[0] + "," + line[1]] || line_by_location[line[1] + "," + line[0]]
}//function isLine